

BEGIN;

-- Creation

CREATE SCHEMA indextest;

CREATE TABLE indextest.customers (
  customerid SERIAL PRIMARY KEY NOT NULL,
  customername text NOT NULL
);

CREATE TABLE indextest.customerclasses (
  classid SERIAL PRIMARY KEY NOT NULL,
  classname text NOT NULL
);

CREATE TABLE indextest.customerclass (
  classid INTEGER NOT NULL REFERENCES indextest.customerclasses(classid),
  customerid INTEGER NOT NULL REFERENCES indextest.customers(customerid)
);

CREATE TABLE indextest.orders (
  orderid SERIAL PRIMARY KEY NOT NULL,
  orderstamp TIMESTAMPTZ NOT NULL DEFAULT now(),
  customerid INTEGER NOT NULL REFERENCES indextest.customers(customerid)
);

CREATE INDEX orders_customerid_orderstamp_idx ON indextest.orders USING btree(customerid,orderstamp);
CREATE INDEX orders_orderstamp_idx ON indextest.orders USING btree(orderstamp);
CREATE INDEX orders_customerid_idx ON indextest.orders USING btree(customerid);

-- Sample Data

-- 10k Customers
INSERT INTO indextest.customers (customerid,customername)
  SELECT s, 'Customer ' || s FROM generate_series(0,9999) as s;

-- 100 Customer Classes
INSERT INTO indextest.customerclasses (classid,classname)
  SELECT s, 'Customer Class ' || s FROM generate_series(0,99) as s;

-- Each Customer into a random Customer Class
INSERT INTO indextest.customerclass (classid,customerid)
  SELECT (random() * 100)::integer % 100, s FROM generate_series(0,9993) as s;
  
-- Customers 9994 to 9999 go into Customer Classes 0 and 1
INSERT INTO indextest.customerclass (classid,customerid)
  SELECT s % 2, s FROM generate_series(9994,9999) as s;

-- 100k Orders with random Customers
INSERT INTO indextest.orders (orderid, orderstamp, customerid)
  SELECT s, now() - (((random() * 100)::integer || ' hour')::interval) - (((random() * 100)::integer % 60 || ' minute')::interval), (random() * 10000)::integer % 10000 FROM generate_series(0,99999) as s;

-- 1k Orders (at end of table) spread across Customers 9994 to 9999 (Customer Classes 0 and 1)
INSERT INTO indextest.orders (orderid, orderstamp, customerid)
  SELECT s, now() - ((s % 1000 + 100 || ' hour')::interval) - (((random() * 100)::integer % 60 || ' minute')::interval), 9994 + (s % 6) FROM generate_series(100000,101000) as s;

-- 200 Customers at end all in class 5 with hundreds of orders each
INSERT INTO indextest.customers (customerid,customername)
  SELECT s, 'Customer ' || s FROM generate_series(10000,10199) as s;
INSERT INTO indextest.customerclass (classid,customerid) 
  SELECT 5, s FROM generate_series(10000,10199) as s;
INSERT INTO indextest.orders (orderid,orderstamp,customerid)
  SELECT s, now() - (((random() * 100)::integer || ' hour')::interval) - (((random() * 100)::integer % 60 || ' minute')::interval), 10000 + (s % 200) FROM generate_series(101001,141000) as s;
  
-- Analyze

ANALYZE indextest.customers;
ANALYZE indextest.customerclasses;
ANALYZE indextest.customerclass;
ANALYZE indextest.orders;

-- Functions to be used

CREATE TYPE indextest.order_and_stamp AS (
  orderid integer,
  orderstamp timestamptz
);
  
CREATE OR REPLACE FUNCTION indextest.orders_by_class(classid_in INTEGER, limit_in INTEGER) RETURNS SETOF indextest.order_and_stamp AS
$$
DECLARE
    customerid_var integer;
    oas_ret indextest.order_and_stamp;
BEGIN
    -- This function performs a separate query for each customer within a customer class for the entire limit
    -- You sort and limit the results returned from this
    FOR customerid_var IN SELECT customerid FROM indextest.customerclass WHERE classid = classid_in LOOP
        FOR oas_ret IN SELECT orderid,orderstamp FROM indextest.orders WHERE customerid = customerid_var ORDER BY orderstamp DESC LIMIT limit_in LOOP
            RETURN NEXT oas_ret;
        END LOOP;
    END LOOP;
    RETURN;
END;
$$ LANGUAGE plpgsql;


CREATE OR REPLACE FUNCTION indextest.orders_by_class_py(classid_in INTEGER, alimit INTEGER) RETURNS SETOF indextest.order_and_stamp AS
$$

# Tables indexed by date, or columns in a table indexed by (attribute, date) are considered as sorted lists; retrieving the first n items
# in a given set of tables ordered by date (or in a set of attributes in a table with a date index on each attribute) can be accomplished by merging
# them in sorted order using a heap. This would be a fast operation if it weren't for the constant costs of writing it this way, which I don't know how to avoid.

from heapq import heappush, heappop, heapify
import time

def totime(tstamp):
     killtz = tstamp.rpartition('-')[0]
     killsecs = killtz.split('.')[0]
     return time.strptime(killsecs, "%Y-%m-%d %H:%M:%S")

# Each sorted (attribute, date) column or table indexed by date is represented by an object that provides a date comparison operator to be used by the heapq implementation.
# Each new row fetched requires a new query into the database (are plpy.execute results cursors, or are they fetched wholesale into python structures? This seemed safest.) Timestamp values 
# are converted to python time structures, meaning additional avoidable costs.

class IterSelect:
     def __init__(self, get_next):
         self.pos = 0
         self.get_next = get_next
 
         self.current = None
         self.date = None
 
         self.gnext()
 
     def gnext(self):
         next = self.get_next(self.pos)
         if next.nrows() == 0:
             raise StopIteration
         else:
             self.pos += 1
             self.current = next[0]
             self.date = totime(self.current['order_date'])
 
     def __cmp__(self, iterselect_inst):
         return cmp(iterselect_inst.date, self.date)

# The heap is used to efficiently merge the tables in sorted order. 

def merge_sorted(gen_heap):
     while len(gen_heap) > 0:
     
         a_gen = heappop(gen_heap)
         try:
             current = a_gen.current
             a_gen.gnext()
             heappush(gen_heap, a_gen)
             yield current
         except StopIteration:
             yield current
 
     raise StopIteration


def make_gen_heap(get_next_list):
     iterselect_list = []
     for get_next in get_next_list:
         try:
             iterselect_list.append(IterSelect(get_next))
         except StopIteration:
             pass
     gen_heap = iterselect_list
     heapify(gen_heap)
 
     return merge_sorted(gen_heap)

get_customers = plpy.prepare("SELECT DISTINCT customerid FROM indextest.customerclass WHERE classid = $1", ["int4"])
get_order_and_date = plpy.prepare("SELECT orderid, orderstamp as order_date FROM indextest.orders WHERE customerid = $1 ORDER BY orderstamp DESC LIMIT 1 OFFSET $2", \
                                    ["int4", "int4"])

def make_get_next(customer):
    return lambda offset: plpy.execute(get_order_and_date, [customer["customerid"], offset])

customers = plpy.execute(get_customers, [classid_in])

get_nexts = [make_get_next(customer) for customer in customers]

gen_heap = make_gen_heap(get_nexts)

# Results are accumulated in a list (could not get generators to work) and returned.

res = []
traversed = 0
for i in gen_heap:
    if traversed == alimit:
        return res
    traversed += 1
    res.append([i['orderid'], i['order_date']])

return res
$$
LANGUAGE plpythonu;

/*  Commented out because this will fill your output with the results


-- Sample Queries

  -- Use double index w/ backwards index scan for single customerid match
  EXPLAIN ANALYZE SELECT orderid FROM indextest.orders WHERE customerid = 9994 ORDER BY orderstamp DESC LIMIT 5;
  EXPLAIN ANALYZE SELECT orderid FROM indextest.orders WHERE customerid = 9994 ORDER BY orderstamp DESC LIMIT 100;
  EXPLAIN ANALYZE SELECT orderid FROM indextest.orders WHERE customerid = 1 ORDER BY orderstamp DESC LIMIT 5;
  EXPLAIN ANALYZE SELECT orderid FROM indextest.orders WHERE customerid = 1 ORDER BY orderstamp DESC LIMIT 100;
  
--  Query Plan for all:
---------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=0.00..14.71 rows=5 width=12) (actual time=0.059..0.125 rows=5 loops=1)
--   ->  Index Scan Backward using orders_customerid_orderstamp_idx on orders  (cost=0.00..585.28 rows=199 width=12) (actual time=0.055..0.112 rows=5 loops=1)
--         Index Cond: (customerid = 9994)
  
--  Runtime is invariant to distribution of customer orders, related only to number of rows returned
--  Average for limit 5 (both densities): 0.227 ms
--  Average for limit 100 (both densities): 0.251 ms / 0.343 ms  (customerid = 1 only has 4 matches)
  
  
----- Methods for multiple customerid match / single customerclass match -----
  
  -- The first two are our most typical use cases.  When pulling even 100 orders (which in the plpgsql function pulls 100 * number of customers in group)
  -- the plpgsql function beats the normal query plan by a significant margin.
  -- Other cases such as where the customer group's orders are dense in a period of time and towards the end of the table, the python function is able to 
  -- beat both other methods.
  
  ----- Sparse/Random Case, Low Limit -----
  --
  
  EXPLAIN ANALYZE
  SELECT o.orderid,o.orderstamp FROM indextest.orders o 
  INNER JOIN indextest.customerclass cc ON (cc.classid = 2) 
  WHERE o.customerid = cc.customerid ORDER BY o.orderstamp DESC LIMIT 5;
  
--  Query Plan:
--  -----------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=350.17..2644.36 rows=5 width=12) (actual time=11.093..106.100 rows=5 loops=1)
--   ->  Nested Loop  (cost=350.17..786798.90 rows=1714 width=12) (actual time=11.090..106.089 rows=5 loops=1)
--         Join Filter: (o.customerid = cc.customerid)
--         ->  Index Scan Backward using orders_orderstamp_idx on orders o  (cost=0.00..15313.61 rows=204004 width=16) (actual time=0.017..1.323 rows=621 loops=1)
--         ->  Materialize  (cost=350.17..351.85 rows=168 width=4) (actual time=0.002..0.084 rows=84 loops=621)
--               ->  Seq Scan on customerclass cc  (cost=0.00..350.00 rows=168 width=4) (actual time=0.429..2.387 rows=84 loops=1)
--                     Filter: (classid = 2)

--  Actual Average Runtime: 10.890 ms
  
  SELECT * FROM indextest.orders_by_class(2,5) ORDER BY orderstamp DESC LIMIT 5;
  
--  Function Average Runtime: 6.201 ms

  SELECT * FROM indextest.orders_by_class_py(2,5);
  
--  Python Average Runtime: 12.119 ms



----- Sparse/Random Case, High Limit -----
  
  EXPLAIN ANALYZE
  SELECT o.orderid,o.orderstamp FROM indextest.orders o 
  INNER JOIN indextest.customerclass cc ON (cc.classid = 2) 
  WHERE o.customerid = cc.customerid ORDER BY o.orderstamp DESC LIMIT 100;
  
--  Query Plan:
--  ---------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=5642.36..5642.61 rows=100 width=12) (actual time=217.213..217.497 rows=100 loops=1)
--   ->  Sort  (cost=5642.36..5646.65 rows=1714 width=12) (actual time=217.210..217.311 rows=100 loops=1)
--         Sort Key: o.orderstamp
--         ->  Hash Join  (cost=352.10..5550.30 rows=1714 width=12) (actual time=8.631..215.774 rows=831 loops=1)
--               Hash Cond: (o.customerid = cc.customerid)
--               ->  Seq Scan on orders o  (cost=0.00..4416.04 rows=204004 width=16) (actual time=5.854..109.622 rows=101001 loops=1)
--               ->  Hash  (cost=350.00..350.00 rows=168 width=4) (actual time=2.452..2.452 rows=84 loops=1)
--                     ->  Seq Scan on customerclass cc  (cost=0.00..350.00 rows=168 width=4) (actual time=0.431..2.342 rows=84 loops=1)
--                           Filter: (classid = 2)

--  Actual Average Runtime: 34.463 ms
  
  SELECT * FROM indextest.orders_by_class(2,100) ORDER BY orderstamp DESC LIMIT 100;

--  Function Average Runtime: 7.847 ms

  SELECT * FROM indextest.orders_by_class_py(2,100);

--  Python Average Runtime: 27.231 ms
 
  
----- Dense/End of Table Case, Low Limit -----

  EXPLAIN ANALYZE 
  SELECT o.orderid,o.orderstamp FROM indextest.orders o 
  INNER JOIN indextest.customerclass cc ON (cc.classid = 0) 
  WHERE o.customerid = cc.customerid ORDER BY o.orderstamp DESC LIMIT 5;
  
--  Query Plan:
--    -----------------------------------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=350.19..2639.74 rows=5 width=12) (actual time=14.159..81.177 rows=5 loops=1)
--   ->  Nested Loop  (cost=350.19..887780.90 rows=1938 width=12) (actual time=14.156..81.166 rows=5 loops=1)
--         Join Filter: (o.customerid = cc.customerid)
--         ->  Index Scan Backward using orders_orderstamp_idx on orders o  (cost=0.00..15313.61 rows=204004 width=16) (actual time=0.018..0.916 rows=417 loops=1)
--         ->  Materialize  (cost=350.19..352.09 rows=190 width=4) (actual time=0.002..0.097 rows=95 loops=417)
--               ->  Seq Scan on customerclass cc  (cost=0.00..350.00 rows=190 width=4) (actual time=0.430..2.322 rows=95 loops=1)
--                     Filter: (classid = 0)

--  Actual Average Runtime: 8.717 ms

  SELECT * FROM indextest.orders_by_class(0,5) ORDER BY orderstamp DESC LIMIT 5;
  
--  Function Average Runtime: 6.695 ms
  
  SELECT * FROM indextest.orders_by_class_py(0,5);
  
--  Python Average Runtime: 13.724 ms



----- Dense/End of Table Case, High Limit -----

  EXPLAIN ANALYZE
  SELECT o.orderid,o.orderstamp FROM indextest.orders o 
  INNER JOIN indextest.customerclass cc ON (cc.classid = 0) 
  WHERE o.customerid = cc.customerid ORDER BY o.orderstamp DESC LIMIT 100;
  
--  Query Plan:
-----------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=5658.63..5658.88 rows=100 width=12) (actual time=219.329..219.610 rows=100 loops=1)
--   ->  Sort  (cost=5658.63..5663.47 rows=1938 width=12) (actual time=219.326..219.422 rows=100 loops=1)
--         Sort Key: o.orderstamp
--         ->  Hash Join  (cost=352.38..5552.81 rows=1938 width=12) (actual time=8.203..216.410 rows=1477 loops=1)
--               Hash Cond: (o.customerid = cc.customerid)
--               ->  Seq Scan on orders o  (cost=0.00..4416.04 rows=204004 width=16) (actual time=5.738..109.593 rows=101001 loops=1)
--               ->  Hash  (cost=350.00..350.00 rows=190 width=4) (actual time=2.415..2.415 rows=95 loops=1)
--                     ->  Seq Scan on customerclass cc  (cost=0.00..350.00 rows=190 width=4) (actual time=0.431..2.293 rows=95 loops=1)
--                           Filter: (classid = 0)

--  Actual Average Runtime: 35.441 ms
  
  SELECT * FROM indextest.orders_by_class(0,100) ORDER BY orderstamp DESC LIMIT 100;

--  Function Average Runtime: 9.563 ms
  
  SELECT * FROM indextest.orders_by_class_py(0,100);
  
--  Python Average Runtime: 28.723 ms
  
  
 
  
----- Many Customers/Orders per Customer Class with small (5) limit -----
  
  EXPLAIN ANALYZE
  SELECT o.orderid,o.orderstamp FROM indextest.orders o 
  INNER JOIN indextest.customerclass cc ON (cc.classid = 5) 
  WHERE o.customerid = cc.customerid ORDER BY o.orderstamp DESC LIMIT 5;
  
--  Query Plan:
-- ----------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=4763.20..4765.70 rows=1000 width=12) (actual time=423.610..426.400 rows=1000 loops=1)
--   ->  Sort  (cost=4763.20..4767.01 rows=1523 width=12) (actual time=423.607..424.565 rows=1000 loops=1)
--         Sort Key: o.orderstamp
--         ->  Hash Join  (cost=352.70..4682.69 rows=1523 width=12) (actual time=3.012..345.063 rows=41099 loops=1)
--               Hash Cond: (o.customerid = cc.customerid)
--               ->  Seq Scan on orders o  (cost=0.00..3786.01 rows=141001 width=16) (actual time=0.012..149.765 rows=141001 loops=1)
--               ->  Hash  (cost=350.00..350.00 rows=216 width=4) (actual time=2.990..2.990 rows=308 loops=1)
--                     ->  Seq Scan on customerclass cc  (cost=0.00..350.00 rows=216 width=4) (actual time=0.013..2.605 rows=308 loops=1)
--                           Filter: (classid = 5)

--  Actual Average Runtime: 2.874 ms
  
  SELECT * FROM indextest.orders_by_class(5,5) ORDER BY orderstamp DESC LIMIT 5;

--  Function Average Runtime: 17.842 ms

  SELECT * FROM indextest.orders_by_class_py(5,5);

--  Python Average Runtime: 38.154 ms
  
  
----- Many Customers/Orders per Customer Class with medium (200) limit -----
  
  EXPLAIN ANALYZE
  SELECT o.orderid,o.orderstamp FROM indextest.orders o 
  INNER JOIN indextest.customerclass cc ON (cc.classid = 5) 
  WHERE o.customerid = cc.customerid ORDER BY o.orderstamp DESC LIMIT 200;
  
--  Query Plan:
------------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=4763.20..4763.70 rows=200 width=12) (actual time=419.899..420.457 rows=200 loops=1)
--   ->  Sort  (cost=4763.20..4767.01 rows=1523 width=12) (actual time=419.897..420.088 rows=200 loops=1)
--         Sort Key: o.orderstamp
--         ->  Hash Join  (cost=352.70..4682.69 rows=1523 width=12) (actual time=3.023..341.457 rows=41099 loops=1)
--               Hash Cond: (o.customerid = cc.customerid)
--               ->  Seq Scan on orders o  (cost=0.00..3786.01 rows=141001 width=16) (actual time=0.012..146.136 rows=141001 loops=1)
--               ->  Hash  (cost=350.00..350.00 rows=216 width=4) (actual time=3.003..3.003 rows=308 loops=1)
--                     ->  Seq Scan on customerclass cc  (cost=0.00..350.00 rows=216 width=4) (actual time=0.012..2.624 rows=308 loops=1)
--                           Filter: (classid = 5)

--  Actual Average Runtime: 93.687 ms
  
  SELECT * FROM indextest.orders_by_class(5,200) ORDER BY orderstamp DESC LIMIT 200;

--  Function Average Runtime: 129.787 ms

  SELECT * FROM indextest.orders_by_class_py(5,200);

--  Python Average Runtime: 71.323 ms
  
  
  
----- Many Customers/Orders per Customer Class with BIG limit -----
  
  EXPLAIN ANALYZE
  SELECT o.orderid,o.orderstamp FROM indextest.orders o 
  INNER JOIN indextest.customerclass cc ON (cc.classid = 5) 
  WHERE o.customerid = cc.customerid ORDER BY o.orderstamp DESC LIMIT 1000;
  
--  Query Plan:
------------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=4763.20..4765.70 rows=1000 width=12) (actual time=423.610..426.400 rows=1000 loops=1)
--   ->  Sort  (cost=4763.20..4767.01 rows=1523 width=12) (actual time=423.607..424.565 rows=1000 loops=1)
--         Sort Key: o.orderstamp
--         ->  Hash Join  (cost=352.70..4682.69 rows=1523 width=12) (actual time=3.012..345.063 rows=41099 loops=1)
--               Hash Cond: (o.customerid = cc.customerid)
--               ->  Seq Scan on orders o  (cost=0.00..3786.01 rows=141001 width=16) (actual time=0.012..149.765 rows=141001 loops=1)
--               ->  Hash  (cost=350.00..350.00 rows=216 width=4) (actual time=2.990..2.990 rows=308 loops=1)
--                     ->  Seq Scan on customerclass cc  (cost=0.00..350.00 rows=216 width=4) (actual time=0.013..2.605 rows=308 loops=1)
--                           Filter: (classid = 5)

--  Actual Average Runtime: 101.939 ms
  
  SELECT * FROM indextest.orders_by_class(5,1000) ORDER BY orderstamp DESC LIMIT 1000;

--  Function Average Runtime: 133.055 ms

  SELECT * FROM indextest.orders_by_class_py(5,1000);

--  Python Average Runtime: 213.113 ms
  
   
   
----- Many Customers/Orders per Customer Class with BIGGER limit -----
  
  EXPLAIN ANALYZE
  SELECT o.orderid,o.orderstamp FROM indextest.orders o 
  INNER JOIN indextest.customerclass cc ON (cc.classid = 5) 
  WHERE o.customerid = cc.customerid ORDER BY o.orderstamp DESC LIMIT 10000;
  
--  Query Plan:
------------------------------------------------------------------------------------------------------------------------------------------
-- Limit  (cost=4763.20..4767.01 rows=1523 width=12) (actual time=419.934..447.761 rows=10000 loops=1)
--   ->  Sort  (cost=4763.20..4767.01 rows=1523 width=12) (actual time=419.931..429.575 rows=10000 loops=1)
--         Sort Key: o.orderstamp
--         ->  Hash Join  (cost=352.70..4682.69 rows=1523 width=12) (actual time=2.971..341.440 rows=41099 loops=1)
--               Hash Cond: (o.customerid = cc.customerid)
--               ->  Seq Scan on orders o  (cost=0.00..3786.01 rows=141001 width=16) (actual time=0.010..145.945 rows=141001 loops=1)
--               ->  Hash  (cost=350.00..350.00 rows=216 width=4) (actual time=2.951..2.951 rows=308 loops=1)
--                     ->  Seq Scan on customerclass cc  (cost=0.00..350.00 rows=216 width=4) (actual time=0.012..2.568 rows=308 loops=1)
--                           Filter: (classid = 5)

--  Actual Average Runtime: 142.393 ms
  
  SELECT * FROM indextest.orders_by_class(5,10000) ORDER BY orderstamp DESC LIMIT 10000;

--  Function Average Runtime: 180.250 ms

  SELECT * FROM indextest.orders_by_class_py(5,10000);

--  Python Average Runtime: 2000 ms

*/

-- Done.
  
DROP SCHEMA indextest CASCADE;

ROLLBACK;


